/*
 *						cc_c
 *						~~~~
 *
 *	 Command line front end for the C68 Compilation system.
 *
 *
 * (c) Copyright 1991-1995 David J. Walker
 *	   Email:  d.j.walker@wins.icl.co.uk
 *
 *	   The purpose of this module is to act as a front end
 *	   to the CPP, C68, AS68 (or QMAC) and LD programs.
 *
 *	   Permission to copy and/or distribute granted under the
 *	   following conditions:
 *
 *	   1). This notice must remain intact.
 *	   2). The author is not responsible for the consequences of use
 *		   this software, no matter how awful, even if they
 *		   arise from defects in it.
 *	   3). Altered version must not be represented as being the
 *		   original software.
 *
 *	This software is based on a program originally written by Jeremy Allison.
 *	It has since been heavily modified.
 *
 *	AMENDMENT HISTORY
 *	~~~~~~~~~~~~~~~~~
 *
 *	Oct 90		JA	v1.0	Original version
 *	31/03/91	DJW v1.1	Made CC print out executed commands if 
 *									verbose option set.
 *							Added -Q option for C68 v3.x
 *	26/05/91	DJW v1.2	Remove output file if a phase fails
 *	27/06/91	DJW v1.3  - Fixed bug whereby incorrect file name given to
 *							LD if starting with _o files.
 *						  - If -TMP specified, then put _o files on this
 *							device, and otherwise only _i and _s files.
 *						  - Proceeding to link assembled files in same run,
 *							then treat _o files as intermediate files
 *	20/06/91	DJW v1.4  - Allowed CPP to be provided its parameters in
 *							both lower case as well as upper case.	 This 
 *							provides better compatibility with LATTICE 
 *							which uses lower case options -i and -d
 *	26/05/91	DJW v1.5  - Added support for -g option to allow programs
 *							called by cc to be off the program path
 *	27/08/91	DJW v2.0  - Amended -tmp and -TMP options to first remove any
 *							existing device name before adding new prefix.
 *						  - The -g option now does a change of the PROG_USE
 *							directory and it is passed through to the daughter
 *							jobs.  This means it affects all references to the
 *							default program directory.
 *						  - A new phase is added to (optionally) be included between
 *							CPP and C68.  This is activated by the -UNPROTO runtime 
 *							parameter to CC.
 *						  - The option to use the QMAC assembler (Quanta GST comaptible
 *							version) as an alternative to the AS68 one supplied with
 *							C68 can be activated using the -QMAC parameter to CC.
 *						 -	Explicit support for -h option removed as no longer
 *							required.  -h will be ignored for backwards compatibility.
 *
 *	19/01/92	DJW v2.01 - The 'no_continue' routine changed to only give the
 *							user a choice if channels have not been passed to CC.
 *						  - Environment variables checked for default values:
 *								TEMP orTMP		location for temporary files.
 *								C68PATH 		path for C68 components.
 *
 *	12/03/92   DJW v2.02 -	Added some checks on directory names for temporary
 *							files and C68 path ending in underscore.
 *						 -	Error routine upgraded to use variable number of
 *							parameters.
 *
 *	23/05/92	DJW v3.00	New version for C68 Release 3.
 *						 -	The -UNPROTO and -U options removed as C68 now
 *							understands ANSI style function prototypes and
 *							declarations.
 *						 -	The -trad option added to stop ANSI extensions
 *							being used.
 *						 -	The environment variables CC_CPP, CC_C68, CC_ASM
 *							and CC_LD can now be used to override the default
 *							program names for each path.
 *
 *	30/06/92   DJW	v3.01 - The C68PATH variable renamed to be CC_PATH.
 *						  - The -v option now passed to C68 and AS68 phases.
 *
 *	20/07/92   DJW	v3.02 - Removed the internal wild-card expansion routine
 *							by switching to generic C68 library option.
 *
 *	24/08/92   DJW	v3.03 - Added implicit '-v' option when cc is invoked
 *							directly without its channels being passed or
 *							redirected.
 *
 *	15/09/92   DJW	v3.04 - Completely rewrote the section on parsing the
 *							command line for options.
 *						  - Added option to automatically run CPP pass on
 *							assembler files whose first character is #
 *							(this is a Unix convention).
 *						  - The -g option replace by -Y option.
 *
 *	21/09/92   DJW	v3.05 - Removed -d option as this is now only likely to
 *							be relevant if running CPP in isolation  (it used
 *							to be treated as a synonym for -D).
 *
 *	22/08/93   DJW	v3.06 - Now pass the -O, -g, -U, -uchar, -error, -frame
 *							and -lattice options through to the C68 program
 *
 *	26 Sep 93  DJW	v4.00 - A full overhaul of the options done to try and
 *							provide the complete list supported by the C68
 *							Release 4 programs.
 *
 *	31 Dec 93  DJW	v4.12 - Added special support for -uchar option.
 *						  - Forced __STDC__ to be undefined when the -trad
 *							option is used.
 *						  - Added support for passing preset options via the
 *							CC_OPTS environment variable.  This involved
 *							re-working the way the command line was handled.
 *	15 Jan 94	DJW v4.20 - Added support for -nostackopt option for C68.
 *						  - Added support for -stackcheck option for C68
 *						  - Added support for -extern option for C68
 *						  - Added support for -trace option for C68
 *						  - Added support for -s option for LD (was same as -S)
 *						  - Added support for -r and -R options for LD RLL use.
 *						  - Added support for -crf option for LD cross-reference
 *						  - Added support for -map option for LD load map
 *						  - Added support for -sym option for LD symbol list
 *						  - Added support for -debug option for LD debug listing
 *
 *	01 Sep 94	DJW v4.21 - Made -nostackopt into dummy as no longer used.
 *						  - Added check that filename follow parameters that
 *							demand this.  Problem report from Franz Herrman.
 *
 *	11 Nov 94	DJW v4.22 - CC_LD environment variable now used.
 *
 *	29 Nov 94	DJW v4.23 - Corrected problem where an environment variable
 *							set to a NULL value was not being treated as if
 *							it was absent.
 *
 *	02 Jan 95	DJW v4.24 - Added use of qopen() routine.
 *						  - Allowed for both . and _ as seperator characters.
 *
 *	12 Mar 95	DJW v4.25 - Added support for -Xa, -Xt and -Xc parameter
 *							options
 *
 *	01 Feb 96	DJW v5.00 - Added support for the George Gwilt assembler GWASS
 *						  - Added checks that GWASS only allowed if 68020+
 *						  - Added checks that -fpu only allowed if FPU present
 *
 *	16 Feb 96	DJW v5.01 - Enhanced support for -r option.
 *						  - Enhanced support for -Xa, -Xt and -Xc options.
 *
 *	03 Mar 95	DJW v5.02 - Extra changes for GWASS support.
 *						  - Removed all calls to STDIO, and replaced with
 *							Unix style I/O to reduce memory footprint.
 *
 *	24 Mar 96	DJW v5.03 - Merged in Jonathan Hudsons XTC68 changes
 */

char _prog_name[] = "CC";
char _version[]   = " v5.02 ";
char _copyright[] = "C68 Compiler driver";
char copyright[]  = "(c)1991-1996  D.J.Walker";


#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#ifdef DOS_LIKE
#include <io.h>
#include <malloc.h>
#endif /* DOS_LIKE */

#ifdef QDOS
#include <unistd.h>
#define __LIBRARY__
#include <qdos.h>
#endif /* QDOS */

#ifdef LIBDEBUG
#include <debug.h>
#else
#define DBG_INIT()
#ifndef JDBG
#define DBG(params)
#else
#define DBG(params) jdbg params
#endif /* JDBG*/
#endif /* LIBDEBUG*/

#ifdef XTC68
#ifdef __unix__
# define DSEP '/'
# define DSEPS "/"
#else
# define DSEP '\\'
# define DSEPS "\\"
#endif /* __unix__ */
#endif /* XTC68 */


/*
 *		Prototypes for routines local to this file
 */

#ifdef __STDC__
#define _P_(params) params
#else
#define _P_(params) ()
#endif

void	add_option			_P_((int, char *));
void	add_tmp 			_P_((char *, char *, char *));
char   *basename			_P_((char *));
int 	command_line		_P_((char *, char *, char *, char *));
void	Environment_Variable _P_((char *, char **,char *));
int 	no_continue 		_P_((void));
void	useage				_P_((void));
struct OPTIONS *what_opt	_P_((char *));

#ifdef QDOS
void (*_consetup)() = consetup_title;			/* Fancy titled screen */
int (*_cmdwildcard)() = cmdexpand;				/* Wild card expansion */
int (*_Open)_P_((const char *, int, ...)) = qopen;
#endif /* QDOS */

#undef _P_

#define STDIN_FILENO	0
#define STDOUT_FILENO	1
#define STDERR_FILENO	2

#define MAX_FNAME 60


struct FILE_LIST {
		struct FILE_LIST *	next;
		char *	name;
		} file_list = {NULL, NULL};

struct cpasses {
	char *p_name;			/* Program name to call for this pass */
	char *p_optbuf; 		/* Pointer to option buffers for this pass */
	char *p_suffix; 		/* Character output file name must end with */
};

int 	pass;							/* Current Pass */
struct cpasses *cp; 					/* Current Pass details */

enum pass_order {
	CC_PASS,  CPP_PASS,   C68_PASS,   AS68_PASS,  QMAC_PASS, GWASS_PASS,
	LD_PASS, LDPRE_PASS, LDPOST_PASS,
	DEL_PASS, COPY_PASS, MAX_PASS
	};
/*
 *	N.B.  This array must correspond to the pass_order above
 */
#ifdef __unix__
# define CC "qcc"
# define CPP "qcpp"
# define LD  "qld"
# define DOT '.'
#else
# define CC "cc"
# define CPP "cpp"
# define LD  "ld"
# ifdef QDOS
#  define DOT '_'
# else
#  define DOT '.'
# endif
#endif


struct cpasses passes[] = {
	{ "CC",   NULL, NULL },
	{ "CPP",  NULL, "i" },
	{ "c68",  NULL, "s" },
	{ "as68", NULL, "o" },          /* Standard C68 assembler */
	{ "qmac", NULL, "o" },          /* Quanta QMAC assembler */
	{ "gwass",NULL, "o" },          /* George GWILT special GWASS assembler */
	{ "LD",   NULL,  0 },
	{ "LD",   NULL,  0 },
	{ "LD",   NULL,  0 },
	{ "rm",   NULL,  0 },
	{ "cp",   NULL,  0 }
};

rllflag = 0;				/* Set if RLL need to be (potentially) linked in */
rlllib	= 0;				/* Set if building RLL library */

enum actions {
		ACTION_IGNORE,					/* Option is not used */
		ACTION_ADD, 					/* Pass option to specified pass */
		ACTION_ADD_REST,				/* Pass option remainder to pass */
		ACTION_ADD_ALL, 				/* Pass option to all passes */
		ACTION_ADD_STKMEM,				/* Stack/Memory option add to all passes */
		ACTION_ADD_MORE,				/* More data follows, allow for space */
		ACTION_ADD_OPTIONAL,			/* More data may follow */
		ACTION_CPP_LAST,				/* Stop after the pre-processor phase */
		ACTION_C68_LAST,				/* Stop after the compiler phase */
		ACTION_ASM_LAST,				/* Stop after the assembler phase */
		ACTION_GWASS,					/* Use GWASS assembler */
		ACTION_QMAC,					/* Use QMAC as assembler phase */
		ACTION_FPU, 					/* Target requires hardware FPU */
		ACTION_VERBOSE, 				/* Run in verbose mode */
		ACTION_VERBOSE_ALL, 			/* Run all components in verbose mode */
		ACTION_TMP, 					/* Set directory for temporary files */
		ACTION_TMP_ALL, 				/* ... temporary and target files */
		ACTION_TRAD,					/* Set -trad option for C68 */
		ACTION_PATH,					/* Set program search path */
		ACTION_RLL, 					/* Set for RLL library to be used */
		ACTION_RLL_LIB, 				/* Set for RLL library being built */
		ACTION_RLL_DEF, 				/* Set for RLL defines used */
		ACTION_UCHAR,					/* Set -uchar options for C68 */
		ACTION_XA,						/* Set -Xa options */
		ACTION_XC,						/* Set -Xc options */
		ACTION_XT,						/* Set -Xt options */
		};

/*
 *	Options list for CC program
 *
 *	This is the complete list of options recognised by the CC program.
 *	For each option, the pass to which it relates is given, plus the
 *	action to be taken on encountering this option.  The table is
 *	searched sequentially, so if options might conflict they must be
 *	put in the table in such an order that this conflict is resolved.
 */
struct OPTIONS {
		char	*identifier;
		int 	pass;
		int 	action;
	} options[] = {
				{"%",       CC_PASS,     ACTION_ADD_STKMEM},
				{"=",       CC_PASS,     ACTION_ADD_STKMEM},
				{"A",       AS68_PASS,   ACTION_ADD},
				{"BUFL",    LDPRE_PASS,  ACTION_ADD},
				{"BUFP",    LDPRE_PASS,  ACTION_ADD},
				{"bufl",    LDPRE_PASS,  ACTION_ADD},
				{"bufp",    LDPRE_PASS,  ACTION_ADD},
				{"CRF",     LDPRE_PASS,  ACTION_ADD},
				{"C",       CPP_PASS,    ACTION_ADD},
				{"crf",     LDPRE_PASS,  ACTION_ADD},
				{"c",       CC_PASS,     ACTION_ASM_LAST},
				{"DEBUG",   LDPRE_PASS,  ACTION_ADD_OPTIONAL},
				{"DRLL_LIBS",CPP_PASS,   ACTION_RLL_DEF},
				{"D",       CPP_PASS,    ACTION_ADD_MORE},
				{"debug",   LDPRE_PASS,  ACTION_ADD_OPTIONAL},
				{"d",       CPP_PASS,    ACTION_ADD},
				{"error",   C68_PASS,    ACTION_ADD},
				{"extern",  C68_PASS,    ACTION_ADD},
				{"E",       CPP_PASS,    ACTION_CPP_LAST},
				{"fpu",     CC_PASS,     ACTION_FPU},
				{"frame",   C68_PASS,    ACTION_ADD},
				{"GWASS",   CC_PASS,     ACTION_GWASS},
				{"gwass",   CC_PASS,     ACTION_GWASS},
				{"g",       C68_PASS,    ACTION_ADD},
				{"H",       CC_PASS,     ACTION_IGNORE},
				{"h",       CC_PASS,     ACTION_IGNORE},
				{"I",       CPP_PASS,    ACTION_ADD_MORE},
				{"icode",   C68_PASS,    ACTION_ADD},
				{"i",       CPP_PASS,    ACTION_ADD_MORE},
				{"L",       LDPRE_PASS,  ACTION_ADD},
				{"i",       CPP_PASS,    ACTION_ADD},
				{"lattice", C68_PASS,    ACTION_ADD},
				{"list",    C68_PASS,    ACTION_ADD},
				{"l",       LDPOST_PASS, ACTION_ADD},
				{"MM",      CPP_PASS,    ACTION_ADD},
				{"M",       CPP_PASS,    ACTION_ADD},
				{"maxerr",  C68_PASS,    ACTION_ADD},
				{"m",       LDPRE_PASS,  ACTION_ADD},
				{"N",       AS68_PASS,   ACTION_ADD},
				{"noformat",C68_PASS,    ACTION_ADD},
				{"noopt",   C68_PASS,    ACTION_ADD},
				{"nopeep",  C68_PASS,    ACTION_ADD},
				{"noreg",   C68_PASS,    ACTION_ADD},
				{"nostdinc",CPP_PASS,    ACTION_ADD},
				{"nostackopt",C68_PASS,  ACTION_IGNORE},
				{"nowarn",  C68_PASS,    ACTION_ADD},
				{"O",       C68_PASS,    ACTION_ADD},
				{"o",       LDPRE_PASS,  ACTION_ADD_MORE},
				{"P",       CPP_PASS,    ACTION_ADD},
				{"pedantic",CPP_PASS,    ACTION_ADD},
				{"peep",    C68_PASS,    ACTION_ADD},
				{"p",       CC_PASS,     ACTION_CPP_LAST},
				{"opt",     C68_PASS,    ACTION_ADD},
				{"QMAC",    CC_PASS,     ACTION_QMAC},
				{"qmac",    CC_PASS,     ACTION_QMAC},
				{"Qtrad",   C68_PASS,    ACTION_TRAD},
				{"Quchar",  C68_PASS,    ACTION_UCHAR},
				{"Q",       C68_PASS,    ACTION_ADD_REST},
				{"reg",     C68_PASS,    ACTION_ADD},
				{"R",       LDPRE_PASS,  ACTION_RLL_LIB},
				{"r",       CC_PASS,     ACTION_RLL},
				{"SYM",     LDPRE_PASS,  ACTION_ADD},
				{"S",       CC_PASS,     ACTION_C68_LAST},
				{"stackcheck",C68_PASS,  ACTION_ADD},
				{"stackopt",C68_PASS,    ACTION_ADD},
				{"sym",     LDPRE_PASS,  ACTION_ADD},
				{"s",       LDPRE_PASS,  ACTION_ADD_MORE},
				{"TMP",     CC_PASS,     ACTION_TMP_ALL},
				{"TRAD",    C68_PASS,    ACTION_TRAD},
				{"tmp",     CC_PASS,     ACTION_TMP},
				{"trace",   C68_PASS,    ACTION_ADD},
				{"trad",    C68_PASS,    ACTION_TRAD},
				{"trigraphs",CPP_PASS,   ACTION_ADD},
				{"UNPROTO", CC_PASS,     ACTION_IGNORE},
				{"U",       CPP_PASS,    ACTION_ADD},
				{"uchar",   C68_PASS,    ACTION_UCHAR},
				{"undef",   CPP_PASS,    ACTION_ADD},
				{"unproto", CC_PASS,     ACTION_IGNORE},
				{"u",       CC_PASS,     ACTION_IGNORE},
				{"U",       CPP_PASS,    ACTION_ADD_MORE},
				{"v",       CC_PASS,     ACTION_VERBOSE_ALL},
				{"V",       CC_PASS,     ACTION_VERBOSE},
				{"warn",    C68_PASS,    ACTION_ADD},
				{"Xa",      CC_PASS,     ACTION_XA},
				{"Xc",      CC_PASS,     ACTION_XC},
				{"Xt",      CC_PASS,     ACTION_XT},
				{"Y",       CC_PASS,     ACTION_PATH},
				{"y",       CC_PASS,     ACTION_PATH},
				{NULL, 0, 0 }
				};

char * tmp_dir = NULL;
int  isname = 0;			/* set as soon as filename found */
int last_pass = AS68_PASS;
int dolink = 1;

int  verbose = 0;
int  alltemp = 0;
int  asmpass = AS68_PASS;	/* Set to identify the assembler to be used */
int  no_c68  = 0;			/* set to skip C68 pass */
#define ASMCPP_OPT	 "-P"	/* Option to CPP for assembler files */

char *ccpath	= NULL;

#ifdef JDBG
jdbg(char *proc, int x, char *ctl,...)
{
	va_list va;
	va_start (va, ctl);

	fputs(proc, stdout);
	fputs(": ", stdout);
	
	vprintf (ctl, va);
	va_end (va);
	fputc ('\n', stdout);
}
#endif /* JDBG */

#ifndef __linux__
#ifndef QDOS
static char *stpcpy(char *d, const char *s)
{
	while(*d++ = *s++);
	return d-1;
}
#endif /* QDOS */
#endif /* __linux__ */

#ifndef QDOS
static char *FilePart(char *fnam)
{
	char *p = fnam;
	char *q = fnam + strlen(fnam);

	for(;q != p; --q)
	{
		if(*q == ':' || *q == DSEP)
		{
			q++;
			break;
		}
	}
	return q;
}

void AddPart(char *d, char *s, int mlen)
{
	int m,n;
	m = strlen(d);
	n = strlen(s);
	if(mlen < m+n+2)
	{
		my_error(5, "Addpart len");
	}
	else
	{
		if(*(d+m-1) != DSEP && *s != DSEP)
		{
			if(*d) strcat(d, DSEPS);
		}
		strcat(d, s); 
	}
}
#endif /* ! QDOS */


int  CheckCPU(void)
{
	return (*((char *)(_sys_var + 0xa1)));
}

/*============================================================ WHAT_OPT */
struct OPTIONS	*what_opt ( opt )
/*				 ~~~~~~~~
 *	Routine to decide if an option is allowable.
 *
 *	Returns NULL on failure, and pointer into options array otherwise.
 *----------------------------------------------------------------------*/
char *opt;
{
	struct OPTIONS *ptr;
	int 	i;

	DBG(("WHAT_OPT",0x18,"Enter: opt='%s'",opt));
	for( i=0, ptr= &options[0] ; ptr->identifier != NULL; ptr++, i++ ) {
		DBG(("WHAT_OPT",0x18,"comparing '%s' with '%s'",opt, ptr->identifier));
		if( !strncmp( opt, ptr->identifier, strlen(ptr->identifier))) {
			DBG(("WHAT_OPT",0x18,"Exit %p",ptr));
			return (ptr);
		}
	}
	DBG(("WHAT_OPT",0x18,"Exit NULL: Option not recognised"));
	return (NULL);
}

/*================================================================= PRINTSTR */
void	printstr ( char * str, ...)
/*		~~~~~~~~
 *	Print a variable number of strings to the error channel.
 *	The array of strings will be NULL terminated.
 *---------------------------------------------------------------------------*/
{
	va_list  ap;

	DBG(("PRINTSTR",0x11,"Enter"));
	va_start(ap, str);
	while(str != NULL) {
		DBG(("PRINTSTR",0x18,"   %s",str));
		(void) write( STDERR_FILENO, str, strlen(str));
		str = (char *)va_arg(ap, char *);
	}
	va_end(ap);
	DBG(("PRINTSTR",0x11,"Exit"));
	return;
}


/*============================================================== NO_CONTINUE */
int 	no_continue ()
/*		~~~~~~~~~~~
 *
 * Routine to check if this compilation set should be aborted.
 *---------------------------------------------------------------------------*/
{
#ifdef QDOS
	extern long _endchanid;
	chanid_t chan = getchid(STDIN_FILENO);
	int  errnosav = errno;
	int  oserrsav = _oserr;
	char c;
	int err;

	if (! _endchanid) 
		return 1;
	printstr ("Compilation failed\nPress <Enter> to exit, any key to continue\n", NULL);
	(void) sd_cure( chan, (timeout_t)(-1)); /* Ensure a cursor is enabled */
	if( (err = io_fbyte( chan, (timeout_t)(-1), &c)) != 0 ) {
		errno = errnosav;
		exit( err );
	}
	errno = errnosav;
	_oserr = oserrsav;
	return ( c == '\n');
#else
	return 0;
#endif /* QDOS */
}


/*================================================================= ADD_TMP */
void	add_tmp (char *outname, char *suffix, char *last_suffix)
/*		~~~~~~~
 *
 *	Routine to compose an output filename with the right extension. 
 *	Does not prepend tmp_dir if suffix is equal to last suffix 
 *	desired.
 *	DJW 07/06/91	unless 'alltemp' set.
 *---------------------------------------------------------------------------*/
{
	char tmpname[MAX_FNAME];
	int extra;

	DBG(("ADD_TMP",0x12,"enter: tmp_dir='%s', outname='%s', suffix='%s', last_suffix='%s'",tmp_dir,outname, suffix, last_suffix));
	(void)strcpy (strrpbrk(outname,"._")+1,suffix);
	DBG(("ADD_TMP",0x18,"new outname = '%s'", outname));
	if (alltemp || strcmp(suffix,last_suffix) ) {
		DBG(("ADD_TMP",0x14,"setting temp directory to '%s'",tmp_dir));
		(void) strcpy( tmpname, tmp_dir);
#ifdef QDOS
		if (isdevice(outname, &extra) && strchr(outname,'_')) {
			DBG(("ADD_TMP",0x14,"removing device name from %s'",outname));
			(void) strcat (tmpname, (char *)((long)strchr(outname, '_')+1));
		} else {
			(void) strcat( tmpname, outname);
		}
#else
		{
			char *p = (char *) FilePart(outname);
			AddPart(tmpname, p, MAX_FNAME);
		}
#endif /* QDOS */
		DBG(("ADD_TMP",0x14," ...giving '%s'",tmpname));
	} else {
		(void) _mkname (tmpname, outname);
	}
	(void) strcpy( outname, tmpname);
	DBG(("ADD_TMP",0x12,"exit: outname($%p)='%s'",outname,outname));
	return;
}

/*================================================================= ADD_OPT */
void	add_option ( pass_opt, p)
/*		~~~~~~~~~~
 *
 *	Add an option to an option buffer.
 *	The buffer will be expanded to make enough room.
 *
 *	Except for the GWASS assembler, a space is
 *	always added before a new option.
 *-------------------------------------------------------------------------*/
int 	pass_opt;
char	*p;
{
	char	**ptr;
	size_t	oldsize;			/* Size of current command line */
	size_t	newbit; 			/* Size of string to be added */
	size_t	spacesize;			/* Count of spaces (if any) as separators */

	DBG(("ADD_OPT",0x14,"pass = %d; p='%s'",pass_opt, p));

	ptr = &passes[pass_opt].p_optbuf;
	DBG(("ADD_OPT",0x18,"   current option buffer='%s'",*ptr));
	DBG(("ADD_OPT",0x18,"   new option to add='%s'",p));
	oldsize = strlen(*ptr);
	newbit = strlen(p);
	spacesize = (oldsize == 0 || pass_opt == GWASS_PASS) ? 0 : 1;

	if ((*ptr = realloc (*ptr, (size_t)(oldsize + newbit + spacesize))) == NULL) {
		printstr("Failure extending option buffers\n", NULL);
		exit (ERR_OM);
	}

	if (spacesize) {
		(void) strcat( *ptr, " ");
	}
	(void) strcat( *ptr, p );
	DBG(("ADD_OPT",0x14,"Exit:  new option buffer='%s'",*ptr));
	return;
}


/*=============================================================== BASENAME */
char   *basename ( fname )
/*		~~~~~~~~
 *
 * Routine to return the base name of any filename. 
 * (Cheats in this implementation as we know we should
 * end in "_ext".  We actually also allow for ".ext".
 *------------------------------------------------------------------------*/
char *fname;
{
	int got_under;

#ifndef QDOS
	return FilePart(fname);
#else
	char *p = &fname[strlen(fname)-1];

	for( got_under = 0; p > fname && got_under < 2; p--) {
		if( strchr("_.",*p)) {
			got_under++;
		}
	}
	return (p == fname ? p : p+2);
#endif /* QDOS */
}


/*=========================================================== COMMAND_LINE */
int 	command_line (p_name, optbuf, inname, outname)
/*		~~~~~~~~~~~~
 *
 *	used to print a command line if verbose specified.	The
 *	strings supplied or printed space seperated.
 *-------------------------------------------------------------------------*/
char	*p_name;
char	*optbuf;
char	*inname;
char	*outname;
{
	int 	fd[2];
	int 	ret;
	char	*cmdbuf;
	size_t	cmdlen;
	char	infullname[50];

	DBG(("COMMAND_LINE",0x14,"p_name='%s' (pass=%d)",p_name,pass));
	DBG(("COMMAND_LINE",0x14,"optbuf/preld   ='%s'",optbuf));
	DBG(("COMMAND_LINE",0x14,"inname/linkfile='%s'",inname));
	DBG(("COMMAND_LINE",0x14,"outname/postld ='%s'",outname));

#ifdef QDOS
	if (pass == GWASS_PASS) {
		fd[0] = 0;
		fd[1] = 1;
		DBG(("COMMAND_LINE",0x14,"GWASS pass, so no output name required"));
		inname = _mkname (infullname, inname);
		DBG(("COMMAND_LINE",0x14,"... and full input name (%s)",inname));
		cmdlen = 0;
	} else {
		cmdlen = strlen(outname) + 2;
	}
#else
	cmdlen = strlen(outname) + 2;
#endif
	cmdlen += strlen (optbuf) + strlen(inname);
	cmdbuf = malloc (cmdlen);

	(void) strcpy (cmdbuf, optbuf);
	switch (pass) {
		case GWASS_PASS:
				break;
		default:
				(void) strcat (cmdbuf, " ");
				break;
	}
	(void) strcat (cmdbuf,inname);
	switch (pass) {
		case GWASS_PASS:
				break;
		default:
				(void) strcat (cmdbuf," ");
				(void) strcat (cmdbuf,outname);
				break;
	}

	if (verbose)
	   (void) printstr("  ",p_name," '",cmdbuf,"'\n", NULL);

	DBG(("COMMAND_LINE",0x14,"about to execute program '%s'",p_name));
	DBG(("COMMAND_LINE",0x14,"... parameters are '%s'",cmdbuf));
#ifdef QDOS
	ret = execl(p_name,pass ==	GWASS_PASS ? fd : (int *)-1, p_name, cmdbuf, NULL);
#else
	{
		char *p, *cmd;
		if(cmd = malloc(4096))
		{
		if(p = getenv("C68PATH"))
		{
			p = stpcpy(cmd, p);
			if(p[-1] != DSEP)
		{
			*p++ = DSEP;
		}
	}
		else
		{
		p = cmd;
	}
		p = stpcpy(p, p_name);
		*p++ = ' ';
		if(inname && *inname)
		{
			p = stpcpy(p, inname);
			*p++ = ' ';
		}
		if(outname && *outname)
		{
			p = stpcpy(p, outname);
			*p++ = ' ';
		}
		strcpy(p, optbuf);
		ret = system(cmd);
		free(cmd);
		}
	}
#endif
	free (cmdbuf);
	DBG(("COMMAND_LINE",0x11,"Exit: ret=%d",ret));
	return (ret);
}


/*==================================================== ENVIRONMENT_VARIABLE */
void	Environment_Variable (name, result, deflt)
/*		~~~~~~~~~~~~~~~~~~~~
 *	This routine is used to read environment variables and set a
 *	global variable appropriately depending on whether the environment
 *	variable was set or not.
 *
 *	N.B.  It is important that the value of an environment variable is
 *		  copied to a new string as changing any environment variable
 *		  can cause the others to be moved.
 *--------------------------------------------------------------------------*/
char	*name;				/* Name of the Environment variable */
char   **result;			/* Location to store result */
char	*deflt; 		  /* Default value if variable not defined */
{
	DBG(("ENVIRONMENT_VARIABLE",0x21,"Enter: looking for %s", name));
	if ((*result = getenv(name)) !=NULL && strlen(*result)) {
		*result=strdup(*result);
	} else {
		*result = deflt;
	}
	DBG(("ENVIRONMENT_VARIABLE",0x21,"Exit: result at %p is '%s'", *result, *result));
	return;
}


/*================================================================== USEAGE */
void	useage()
/*		~~~~~~
 *	Simply print a usage message.
/*--------------------------------------------------------------------------*/
{
	(void) printstr ("useage: cc [options] filename[s]\n",NULL);
	exit( -1 );
}


/*============================================================== PARAMETERS */
void	Parameters (int argc, char * argv[])
/*		~~~~~~~~~~
 *	Process a set of parameters.  These may have been passed via the
 *	command line, or via an environment variable - we do not care.
 *--------------------------------------------------------------------------*/
{
	char workbuf[100];
	char *argvsf;
	int sf;

	DBG(("PARAMETERS",0x11,"Enter: argc=%d", argc));
	for( sf = 1; sf < argc; sf++) {
		struct OPTIONS *optptr;

		DBG(("PARAMETERS",0x12,"parameter parsing 'sf' loop - argv[sf]='%s'", \
										argv[sf]));
		if( *argv[sf] != '-' ) {
			struct FILE_LIST *ptr;
			DBG(("PARAMETERS",0x12,"filename - adding to list"));
			isname++;
			for (ptr = &file_list ; ptr->next ; ptr=ptr->next)
				;
			DBG(("PARAMETERS",0x24,"file_list starts at %p, finishes at %p", \
				  &file_list, ptr));
			ptr->next = malloc (sizeof(struct FILE_LIST));
			DBG (("PARAMETERS",0x24,"Added new FILE_LIST structure at address %p", \
											ptr->next));
			if ((ptr=ptr->next) == NULL)
			{
				DBG(("PARAMETERS",0x21,"Failed to allocate FILE_LIST structure"));
				exit (ERR_OM);
			}
			ptr->next = NULL;
			ptr->name = argv[sf];
			continue; /* Skip filenames */
		}

		/* Stop after getting options */
		argvsf = argv[sf];

		/*
		 *	Identify the option.
		 *	The value of returned as an index into
		 *	the option array.
		 */
		DBG (("PARAMETERS",0x28,"last_pass = %d",last_pass));
		if ((optptr = what_opt( &argvsf[1] )) == NULL ) {
			DBG(("PARAMETERS",0x12,"   option not recognised - so ignored"));
			printstr ("option '",argvsf,"' not recognised - ignored\n",NULL);
			continue;
		}
		DBG(("PARAMETERS",0x12," action=%d, pass=%d",optptr->action,optptr->pass));
		switch (optptr->action) {
		case ACTION_ADD_STKMEM:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_STKMEM)"));
				argvsf++;		/* Skip over - character */
				add_option (CPP_PASS, argvsf);
				add_option (C68_PASS, argvsf);
				add_option (AS68_PASS,argvsf);
				add_option (LD_PASS,  argvsf);
				break;				

		case ACTION_ADD_ALL:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_ALL)"));
				add_option (CPP_PASS, argvsf);
				add_option (C68_PASS, argvsf);
				add_option (AS68_PASS,argvsf);
				add_option (LD_PASS,  argvsf);
				break;				

		case ACTION_ADD:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD)"));
				add_option (optptr->pass,argvsf);
				break;

		case ACTION_ADD_REST:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_REST)"));
				(void) strcpy(workbuf,"-");
				add_option (optptr->pass, strcat(workbuf, argvsf+2));
				break;

		case ACTION_ADD_MORE:
ADD_MORE:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_MORE)"));
				/*
				 *	There must be more (typically a filename).
				 *	If necessary get the next parameter
				 */
				if (argvsf[2] == '\0')
				{
					(void) strcpy (workbuf,argvsf);
					argvsf = argv[++sf];
					(void) strcat(workbuf, argvsf);
					add_option (optptr->pass, workbuf);
					break;
				}
				add_option (optptr->pass,argvsf);
				break;

		case ACTION_ADD_OPTIONAL:
				DBG(("PARAMETERS",0x12,"   (ACTION_ADD_OPTIONAL)"));
				add_option (optptr->pass,argvsf);
				if (*(argv[sf+1]) != '-') {
					add_option (optptr->pass, argv[++sf]);
				}
				break;

		case ACTION_IGNORE:
				DBG(("PARAMETERS",0x12,"   (ACTION_IGNORE)"));
				break;

		case ACTION_TMP_ALL:
				DBG(("PARAMETERS",0x12,"   (ACTION_TMP_ALL)"));
				alltemp = 1;
				/* FALLTHRU */
		case ACTION_TMP: /* Define temp directory */
				DBG(("PARAMETERS",0x12,"   (ACTION_TMP)"));
				/* Ensure directory ends in '_' */
				if( argvsf[strlen(argvsf)-1] != '_' )
#ifdef QDOS
				{
					printstr("Temporary directory (", argvsf, ") must end in _\n", NULL);
					exit (ERR_BP);
				}
#endif /* QDOS */
				tmp_dir = &argvsf[4];
				DBG(("PARAMETERS",0x12,"   tmp_dir=%s",tmp_dir));
				break;

		case ACTION_VERBOSE_ALL:
				DBG(("PARAMETERS",0x12,"   (ACTION_VERBOSE_ALL)"));
				verbose++;
				add_option (CPP_PASS, argvsf);
				add_option (C68_PASS, argvsf);
				add_option (AS68_PASS,argvsf);
				add_option (LD_PASS,  argvsf);
				break;				

		case ACTION_VERBOSE:
				DBG(("PARAMETERS",0x12,"   (ACTION_VERBOSE)"));
				verbose++;
				break;

		case ACTION_CPP_LAST:
				DBG(("PARAMETERS",0x12,"   (ACTION_CPP_LAST)"));
				add_option (CPP_PASS, argvsf);
				last_pass = CPP_PASS;
				dolink = 0;
				break;

		case ACTION_C68_LAST:
				DBG(("PARAMETERS",0x12,"   (ACTION_C68_LAST)"));
				last_pass = C68_PASS;
				dolink = 0;
				break;

		case ACTION_ASM_LAST:
				DBG(("PARAMETERS",0x12,"   (ACTION_ASM_LAST)"));
				last_pass = AS68_PASS;
				dolink = 0;
				break;

		case ACTION_TRAD:
				DBG(("PARAMETERS",0x12,"   (ACTION_TRAD)"));
				add_option( CPP_PASS, "-U__STDC__");
				add_option (optptr->pass,argvsf);
				break;

#ifdef QDOS
		case ACTION_FPU:
				DBG(("PARAMETERS",0x12,"   (ACTION_FPU)"));
				add_option (C68_PASS,"-fpu");
				add_option (LDPRE_PASS,"-lfpu");
				if (CheckCPU() < 0x20) {
					printstr ("The '-fpu' option needs the GWASS assembler\n", NULL);
					printstr ("The GWASS assembler requires a 68020 or better\n", NULL);
					exit (-1);
				}
				/* FALLTHRU */

		case ACTION_GWASS:
				DBG(("PARAMETERS",0x12,"   (ACTION_GWASS)"));
				if (CheckCPU() < 0x20) {
					printstr ("The GWASS assembler requires a 68020 or better\n", NULL);
					exit(-1);
				}
				add_option (C68_PASS,"-qmc68k");
				add_option (GWASS_PASS,"D/5/Q//A/1/");
				asmpass = GWASS_PASS;
				passes[C68_PASS].p_suffix = "asm";
				break;

		case ACTION_QMAC:
				DBG(("PARAMETERS",0x12,"   (ACTION_QMAC)"));
				asmpass = QMAC_PASS;
				passes[C68_PASS].p_suffix = "asm";
				break;
#endif /* QDOS */

		case ACTION_PATH:
				DBG(("PARAMETERS",0x12,"   (ACTION_PATH)"));
				ccpath = &argvsf[2];
				break;

		case ACTION_RLL_DEF:
				DBG(("PARAMETERS",0x12,"    (ACTION_RLL_DEF)"));
				rllflag++;
				goto ADD_MORE;

		case ACTION_RLL_LIB:
				DBG(("PARAMETERS",0x12,"    (ACTION_RLL_LIB)"));
				rlllib++;
				goto ADD_MORE;

		case ACTION_RLL:
				DBG(("PARAMETERS",0x12,"    (ACTION_RLL)"));
				add_option (CPP_PASS, "-DRLL_LIBS");
				rllflag++;
				goto ADD_MORE;

		case ACTION_UCHAR:
				DBG(("PARAMETERS",0x12,"   (ACTION_UCHAR)"));
				add_option (CPP_PASS,"-D__CHAR_UNSIGNED__");
				add_option (optptr->pass,argvsf);
				break;
		case ACTION_XA:
				add_option (LDPRE_PASS, "-lxa");
				add_option (CPP_PASS, "-D__STDC__=0");
				add_option (CPP_PASS, "-trigraphs");
				break;
		case ACTION_XC:
				add_option (C68_PASS, "-obsolete");
				add_option (CPP_PASS, "-pedantic");
				add_option (CPP_PASS, "-D__STRICT_ANSI__= ");
				add_option (LDPRE_PASS, "-lxc");
				add_option (CPP_PASS, "-D__STDC__=1");
				add_option (CPP_PASS, "-trigraphs");
				break;
		case ACTION_XT:
				add_option (LDPRE_PASS, "-lxt");
				add_option (CPP_PASS, "-D__STDC__=0");
				break;
		default:
				DBG(("PARAMETERS",0x12,"   (default)"));
				useage();
				break;
		} /* end of 'index' switch */
		/*
		 *	Option has been eaten - mark space unused
		 */
		argv[sf] = NULL;
	}	/* end of 'options' for loop */
	return;
}


/*================================================== ENVIRONMENT_AORAMETERS */
static
void	Environment_Parameters (char * argv[])
/*		~~~~~~~~~~~~~~~~~~~~~~
 *	This routine is responsible for getting the CC_OPTS environment
 *	variable, and then using it to set up data structures that look
 *	like the 'argc' and 'argv' ones passed to main(), before calling
 *	the shared parameter analysis routine.
 *-------------------------------------------------------------------------*/
{
	char	*cc_opts;
	int 	envc;
	char ** envv;

	Environment_Variable ("CC_OPTS", &cc_opts, "");
	/*
	 *	Initialise vector and count
	 */
	envc = 1;
	envv = (char **)malloc (2 * sizeof(char *));
	envv[0] = argv[0];
	envv[1] = NULL;

	/*
	 *	Now parse the Environment Variable, splitting it
	 *	up into NULL terminated strings and building the
	 *	vector array.
	 */
	for ( ;  ; ) {
		char sc;
		/*
		 *	If we have any leading whitespace characters
		 *	then we need to skip over them.
		 */
		while (*cc_opts == ' ') {
			cc_opts++;
		}
		if (*cc_opts == '\0') {
			break;
		}
		sc = ' ';
		/*
		 *	Look for any of the characters which are
		 *	stored, but treated specially.
		 */
		switch (*cc_opts) {
		case '\'' :
		case '"'  :
				/* Next argument is string - store terminator */
				sc = *cc_opts++;
				break;
		case '\\' :
				cc_opts++;
				break;
		}
		/*
		 *	There is at least one more element, so add this
		 *	new element to the vector array.
		 */
		envv = (char **)realloc (envv, (size_t)((envc + 2) * sizeof (char *)));
		envv[envc++] = cc_opts;
		envv[envc] = NULL;
		/*
		 * Parse until terminating character reached
		 * The terminating character is changed to a NULL byte.
		 * Allow for premature termination by NULL byte.
		 */
		for ( ;  ; cc_opts++ ) {
			if (*cc_opts == '\0') {
				break;
			}
			if (*cc_opts == sc) {
				*cc_opts++ = '\0';
				break;
			}
		}
	} /* End of main for loop */

	Parameters (envc,envv);
	return;
}

 
/*==================================================================== MAIN */
void	main ( argc, argv)
/*		~~~~
 *
 * Main routine to do the compilations, managing all arguments. 										  *
 *--------------------------------------------------------------------------*/
int argc;
char *argv[];
{
	struct FILE_LIST *nameptr;
	char inname[MAX_FNAME], outname[MAX_FNAME];
	char *p, *p1;
	int 	n;
	int 	ret;

	DBG_INIT();
#ifdef QDOS
	if (iscon(getchid(1),0) && !isnoclose(1)) {
		verbose++;
	}
#endif /* QDOS */
	/*
	 *	Initialise options buffers
	 */
	for (n = 0 ; n < MAX_PASS ; n++) {
		if ((p = malloc((size_t)1)) == NULL) {
			exit (ERR_OM);
		}
		*p = '\0';
		passes[n].p_optbuf = p;
	}
	/*
	 *	Preset any values that will be got
	 *	from Environment variables.
	 */
#ifdef QDOS
	Environment_Variable ("TMP",&tmp_dir, "");
	if (!strcmp(tmp_dir,"")) {
		Environment_Variable ("TEMP",&tmp_dir, "");
	}
#endif /* QDOS */

	Environment_Variable ("CC_PATH",&ccpath, "");
	Environment_Variable ("CC_CPP", &passes[CPP_PASS].p_name, "cpp");
	Environment_Variable ("CC_C68", &passes[C68_PASS].p_name, "c68");
	Environment_Variable ("CC_ASM", &passes[AS68_PASS].p_name,"as68");
	Environment_Variable ("CC_LD",	&passes[LD_PASS].p_name,  "ld");

	/*
	 *	Handle parameters.
	 */
	Environment_Parameters(argv);
	Parameters (argc, argv);

	DBG (("MAIN",0x21,"Finished parsing parameters"));

#ifdef QDOS
	if (verbose) {
		DBG(("MAIN",0x14,"CC running in VERBOSE mode"));
		printstr (" ", _prog_name, _version, " (",_copyright, ") ", copyright, "\n", NULL);
	}
#endif /* QDOS */
	/*
	 *	Now do some consistency checks
	 */
	DBG(("MAIN",0x18,"Checking that isname set (=%d)",isname));
	if( isname == 0 ) {
		DBG(("MAIN",0x14,"No filenames supplied to compile"));
		printstr("No filenames to compile\n", NULL);
		exit (ERR_BP);
	}

#ifdef QDOS
	DBG(("MAIN",0x18,"Checking tmp_dir - %p (%s)", tmp_dir, tmp_dir));
	if ( (n=strlen(tmp_dir)) != 0 ) {
		DBG(("MAIN",0x18,"tmp_dir set - checking ends in underscore"));
		if (tmp_dir[n-1] != '_' ) {
			printstr("Directory for temporary files (",tmp_dir,") must end in _\n",NULL);
			exit (ERR_BP);
		}
	}
#endif /* QDOS */

	DBG(("MAIN",0x18,"Checking ccpath - %p (%s)", ccpath, ccpath));
	if (strlen(ccpath)) {
		DBG(("MAIN",0x18,"ccpath set - checking ends in underscore"));
		if (ccpath[strlen(ccpath)-1] != '_' ) {
			DBG(("MAIN",0x18,"... failed (%s)",ccpath));
			printstr("Directory for C68 phases (",ccpath,") must end in _\n", NULL);
			exit (ERR_BP);
		}
#ifdef QDOS
		DBG(("MAIN",0x18,"ccpath set - need to set as program directory"));
		if (chpdir (ccpath)) {
			DBG(("MAIN",0x18,"... failed"));
			printstr("Failed to set Directory Path for C68 phases to ",ccpath,"\n",NULL);
			exit (ERR_BP);
		}
#endif /* QDOS */
	}

	if (last_pass == AS68_PASS) {
		DBG(("MAIN",0x18,"assembler is the last pass"));
		DBG(("MAIN",0x18,"... so ensure correct assembler used"));
		last_pass = asmpass;
	}
	if (rllflag)
	{
		add_option (LDPOST_PASS,"-rc");
	}
	/*
	 *	We have now finished handling all the parameters,
	 *	so it is time to start compiling programs
	 */
	DBG(("MAIN",0x14,"Start compilation phase"));
	DBG(("MAIN",0x18,"... first FILE_LIST structure at %p", file_list.next));
	for( nameptr = file_list.next; nameptr; nameptr = nameptr->next) {
		int 	start_pass;
		DBG(("MAIN",0x14,"Next file to process is '%s'",nameptr->name));
		(void) strcpy( inname, nameptr->name ); /* set initial input filename */
#ifdef QDOS
		p1 = &inname[strlen(inname)-2]; /* Point at the end of the input name */
		if (*p1 == '.') {               /* Is seperator a fullstop ? */
			*p1 = '_';					/* ... yes then change to underscore */
		}
		if(*p1++ != '_') {
#else
		p1 = inname + strlen(inname) - 2;
		if(*p1++ != '.') {
#endif
			printstr("Filename '",inname,"' has no extension\n", NULL);
			exit (ERR_BP);
		}
		switch( toupper((unsigned char)*p1) ) {
		case 'C':
				DBG(("MAIN",0x18,"File is a _C file"));
				start_pass = CPP_PASS;
				break;
		case 'I':
				DBG(("MAIN",0x18,"File is a _I file"));
				start_pass = C68_PASS;
				break;
		case 'S':
				DBG(("MAIN",0x18,"File is a _S file"));
ASM_PASS:
				start_pass = asmpass;
				if (asmpass == AS68_PASS) {
					int  fd;
					char ch;
					DBG(("MAIN",0x18,"...using AS68 assmebler"));
					/*
					 *	This is to allow for the special case of an
					 *	assembler file that starts with a # character
					 *	as this means it needs to be pre-processed
					 */
					if ((fd = open(inname,O_RDONLY)) > 0) {
						(void) read (fd, &ch, 1);
						(void) close (fd);
						if (ch == '#') {
							no_c68 = 1;
							start_pass = CPP_PASS;
							add_option (CPP_PASS, ASMCPP_OPT);
						}
					}
				}
				break;
		case 'X':
				DBG(("MAIN",0x18,"File is a _X file"));
				no_c68 = 1;
				start_pass = CPP_PASS;
				add_option (CPP_PASS, ASMCPP_OPT);
				passes[CPP_PASS].p_suffix = (asmpass == AS68_PASS) ? "o" : "asm";
				break;
		case 'O':
				DBG(("MAIN",0x18,"File is a _O file"));
				start_pass = LD_PASS;
				add_option( LD_PASS, inname);
				break;
		case 'A':
				DBG(("MAIN",0x18,"File might be an _ASM file"));
				if (!stricmp(p1,"asm")) {
					goto ASM_PASS;
				}
				/* FALL THRU */
		default:
			printstr("Unknown extension '_",*p1,"'\n",NULL);
			exit (ERR_BP);
		}
		DBG (("MAIN",0x18,"start_pass=%d, last_pass=%d",start_pass,last_pass)); 
		for(pass=start_pass ; (pass <= last_pass) && (pass < LD_PASS) ; pass++) {
			/*
			 *	This is to allow for the fact that _s files
			 *	that start with # in the first character are
			 *	put through the pre-processor first
			 */
			if ((pass == C68_PASS && no_c68)
			||	(pass == AS68_PASS) ) {
				pass = asmpass;
			}

			cp = &passes[pass];
			if (pass != last_pass)
				(void) strcpy( outname, basename( inname )); /* Get its basename */
			else
				(void) strcpy( outname, nameptr->name);    /* get full original name */

			/* Add the temp)orary directory, and ensure suffix is correct */
			add_tmp( outname, cp->p_suffix, passes[last_pass].p_suffix);
			ret=command_line (cp->p_name, cp->p_optbuf, inname, outname);
			if ((ret != 0)
			&& !(ret == -1 && pass == GWASS_PASS)) /* Patch for GWASS */
				{
				if (pass == GWASS_PASS) {
					char  errbuf[10];
					(void) itoa (ret, errbuf);
					printstr ("GWASS returned error code ",errbuf,"\n",NULL);
				}
				DBG(("MAIN",0x18,"command failed (%d) - tidying up",ret));
				(void) unlink (outname);   /* remove failed output file */
				if( no_continue()) {
					exit( ret );
				}
				dolink = 0; /* A pass failed - forget linking */
				break;
			}
			/*
			 *	Delete any intermediate files
			 *	- but not _o's
			 */
			if( pass != start_pass ) {
				DBG(("MAIN",0x18,"Deleting intermediate file '%s'",inname));
				(void) unlink( inname );
			}
			if (pass == GWASS_PASS) {
				pass = COPY_PASS;
				(void) strcpy(strrpbrk (inname,"._"),"_o");
				if (strcmp(inname,outname)) {
					(void) command_line ("cp", "", inname, outname);
					(void) unlink (inname);
				}
				(void) strcpy(strrpbrk (inname,"._"),"_LST");
				DBG(("MAIN",0x18,"Deleting GWASS list file '%s'",inname));
				(void) unlink (inname);
				(void) strcpy(strrpbrk (inname,"._"),"_SYM");
				DBG(("MAIN",0x18,"Deleting GWASS symbol file '%s'",inname));
				(void) unlink (inname);
				pass = GWASS_PASS;
			}
			(void) strcpy( inname, outname); /* Output from previous pass goes into
									  next pass */
		}

		if( dolink && start_pass != LD_PASS ) {
			DBG(("MAIN",0x18,"adding file %s to list of link files",outname));
			add_option( LD_PASS, outname);
			add_option( DEL_PASS, outname);
		}
	}

	DBG(("MAIN",0x18,"Is link required?  dolink=%d",dolink));
	if(!dolink ) {
		exit (ret);
	}

	/*
	 *	We have now finished the compilation phase
	 *	Finally do the link if neccessary.
	 */
	DBG(("MAIN",0x14,"Start link phase"));
	if (! (ret = command_line (passes[LD_PASS].p_name,
							   passes[LDPRE_PASS].p_optbuf,
							   passes[LD_PASS].p_optbuf,
							   passes[LDPOST_PASS].p_optbuf))) {
		/*
		 *	delete any intermediate object 
		 *	files at this point
		 */
		for (p=passes[DEL_PASS].p_optbuf ; *p != 0 ; p += strlen(p) + 1 ) {
			while (*p == ' ') p++;	/* skip any leading spaces */
			if ( (p1 = strpbrk (p, " ")) != NULL) {
				*p1 = '\0'; 	/* Convert space to terminator */
			}
			(void) unlink (p);
		}
	} else {
		printstr("Link failed\n", NULL);
		exit(ret);
	}
	DBG(("MAIN",0x11,"CC finished - return code = %d",ret));
	exit (ret);
}
